import gi

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk, Gio, GLib, GdkPixbuf  # noqa: E402,F401
gi.require_version('Notify', '0.7')
from gi.repository import Notify  # noqa: E402,F401
